use crate::api::PlayStoneRequest;
use serde::{Deserialize, Serialize};
use std::time::Duration;
use typeshare::typeshare;

#[derive(Clone, Copy, Debug, PartialEq, Eq, Serialize, Deserialize)]
#[typeshare]
pub enum Color {
    Black,
    White,
}

#[derive(Clone, Copy, Debug, Serialize, Deserialize)]
#[typeshare]
pub struct Size {
    pub width: u8,
    pub height: u8,
}

impl Default for Size {
    fn default() -> Self {
        Self {
            width: 19,
            height: 19,
        }
    }
}

#[derive(Debug)]
pub struct AppState {
    pub game: Option<GameState>,
}

impl AppState {
    pub fn new() -> Self {
        Self {
            game: Some(GameState::new()),
        }
    }

    pub fn place_stone(&mut self, req: PlayStoneRequest) {
        match self.game {
            Some(ref mut game) => {
                game.place_stone(req.column, req.row);
            }
            None => {}
        }
    }
}

#[derive(Debug)]
enum Rank {
    Kyu(u8),
    Dan(u8),
    Pro(u8),
}

impl TryFrom<&str> for Rank {
    type Error = String;

    fn try_from(_: &str) -> Result<Rank, Self::Error> {
        Ok(Rank::Kyu(15))
    }
}

impl From<Rank> for String {
    fn from(r: Rank) -> String {
        match r {
            Rank::Kyu(v) => format!("{} kyu", v),
            Rank::Dan(v) => format!("{} dan", v),
            Rank::Pro(v) => format!("{} pro", v),
        }
    }
}

#[derive(Debug)]
pub struct Player {
    name: String,
    rank: Rank,
}

#[derive(Debug)]
pub struct GameState {
    pub board: Board,
    pub conversation: Vec<String>,
    pub current_player: Color,

    pub white_player: Player,
    pub black_player: Player,

    pub white_clock: Duration,
    pub black_clock: Duration,
}

impl GameState {
    fn new() -> GameState {
        GameState {
            board: Board::new(),
            conversation: vec![],
            current_player: Color::Black,
            white_player: Player {
                name: "Savanni".to_owned(),
                rank: Rank::Kyu(10),
            },
            black_player: Player {
                name: "Opal".to_owned(),
                rank: Rank::Kyu(10),
            },
            white_clock: Duration::from_secs(600),
            black_clock: Duration::from_secs(600),
        }
    }

    fn place_stone(&mut self, column: u8, row: u8) {
        self.board.place_stone(column, row, self.current_player);
        match self.current_player {
            Color::White => self.current_player = Color::Black,
            Color::Black => self.current_player = Color::White,
        }
    }
}

#[derive(Debug)]
pub struct Board {
    pub size: Size,
    pub spaces: Vec<Option<Color>>,
}

impl Board {
    fn new() -> Self {
        let mut spaces = Vec::new();
        for _ in 0..19 * 19 {
            spaces.push(None);
        }
        Self {
            size: Size {
                width: 19,
                height: 19,
            },
            spaces,
        }
    }

    pub fn place_stone(&mut self, column: u8, row: u8, stone: Color) {
        let addr = self.addr(column, row);
        self.spaces[addr] = Some(stone);
    }

    pub fn stone(&self, column: u8, row: u8) -> Option<Color> {
        let addr = self.addr(column, row);
        self.spaces[addr]
    }

    fn addr(&self, column: u8, row: u8) -> usize {
        ((row as usize) * (self.size.width as usize) + (column as usize)) as usize
    }
}
